#include <sys/types.h>

#include <iomanip>
#include <iostream>

#include <boost/filesystem.hpp>

namespace fs = boost::filesystem;

extern char	*__progname;

static off_t	dirsize(const fs::path &, unsigned *, unsigned *);

static off_t
dirsize(const fs::path &path_dir, unsigned *out_dirs, unsigned *out_files)
{
	off_t tot;
	unsigned dirs;
	unsigned files;

	*out_dirs = 0;
	*out_files = 0;

	tot = 0;
	fs::directory_iterator end;
	for (fs::directory_iterator iter(path_dir); iter != end; ++iter) {
		if (fs::is_symlink(iter->path())) continue;
		if (fs::is_directory(iter->path())) {
			tot += dirsize(iter->path(), &dirs, &files);
			*out_dirs += dirs + 1;
			*out_files += files;
		} else {
			try {
				tot += fs::file_size(iter->path());
				++*out_files;
			} catch (const fs::filesystem_error &e) {
				std::cerr << __progname << ": "
				    << e.what() << '\n';
			}
		}
	}

	return tot;
}

int
main(int argc, char **argv)
{
	off_t size;
	unsigned dirs;
	unsigned files;
	unsigned tot_dirs;
	unsigned tot_files;

	if (argc < 2) {
		std::cout << "usage: " << __progname
		    << " ( FILE | DIR )...\n";
		return 1;
	}

	size = 0;
	tot_dirs = 0;
	tot_files = 0;
	for (argv++; *argv; argv++) {
		fs::path path(*argv);
		if (is_directory(path)) {
			size += dirsize(path, &dirs, &files);
			tot_dirs += 1 + dirs;
			tot_files += files;
		} else {
			try {
				size += fs::file_size(path);
				++tot_files;
			} catch (const fs::filesystem_error &e) {
				std::cerr << __progname << ": "
				    << e.what() << '\n';
			}
		}
	}

	if (!tot_dirs && !tot_files) {
		std::cerr << __progname << ": no files or directories\n";
		return 1;
	}

	std::cout << "dirs:  " << tot_dirs
	    << "\nfiles: " << tot_files
	    << "\nbytes: " << size;

	std::cout.fill('0');
	if (size >> 40) {
		// 64 exobytes is enough for anybody
		uint16_t whole = size >> 40;
		uint32_t remainder = size & 0x3fffffffLL;
		float frac = remainder /
		    (1024.0f * 1024.0f * 1024.0f * 1024.0f);
		std::cout << " (" << whole << '.'
		    << std::setw(2) << static_cast<int>(frac * 100 + .5)
		    << " TB)";
	} else if (size >> 30) {
		uint16_t whole = size >> 30;
		uint32_t remainder = size & 0x3fffffff;
		float frac = remainder / (1024.0f * 1024.0f * 1024.0f);
		std::cout << " (" << whole << '.'
		    << std::setw(2) << static_cast<int>(frac * 100 + .5)
		    << " GB)";
	} else if (size >> 20) {
		uint16_t whole = size >> 20;
		uint32_t remainder = size & 0xfffff;
		float frac = remainder / (1024.0f * 1024.0f);
		std::cout << " (" << whole << '.'
		    << std::setw(2) << static_cast<int>(frac * 100 + .5)
		    << " MB)";
	} else if (size >> 10) {
		uint16_t whole = size >> 10;
		uint16_t remainder = size & 0x3ff;
		float frac = remainder / 1024.0f;
		std::cout << " (" << whole << '.'
		    << std::setw(2) << static_cast<int>(frac * 100 + .5)
		    << " KB)";
	}
	std::cout << '\n';
	return 0;
}
